﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Runtime.CompilerServices;
using System.Text.RegularExpressions;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
//using System.Windows.Shapes;

namespace DreamcastGdiTexture
{
    /// <summary>
    /// Interaction logic for MainWindow.xaml
    /// </summary>
    public partial class MainWindow : Window, INotifyPropertyChanged
    {
        private string currentAppPath;
        private string gditoolsPath;
        private string dospvrPath;
        private string pvr2pngPath;
        //private string buildgdiPath;

        private const string gdiFilter = "*.gdi";

        private const string DATAPATH = "DATA";
        private const string IPBINPATH = @"BOOT\IP.BIN";
        private const string REBUILDPATH = @"REBUILD";
        private const string GDTEXFILENAME = "0GDTEX.PVR";

        private DiscUtils.Gdrom.GDromBuilder builder;

        private bool isAborting = false;

        public ObservableCollection<GDIItem> ItemList { get; private set; } = new ObservableCollection<GDIItem>();

        private bool _IsBusy;
        public bool IsBusy
        {
            get { return _IsBusy; }
            set { _IsBusy = value; raiseEvent(); }
        }

        private string _TempFolder;
        public string TempFolder
        {
            get { return _TempFolder; }
            set { _TempFolder = value; raiseEvent(); }
        }

        public event PropertyChangedEventHandler PropertyChanged;

        public MainWindow()
        {
            InitializeComponent();

            currentAppPath = AppDomain.CurrentDomain.BaseDirectory;
            gditoolsPath = Path.Combine(currentAppPath, "tools", "gditools.exe");
            //buildgdiPath = Path.Combine(currentAppPath, "tools", "buildgdi.exe");
            dospvrPath = Path.Combine(currentAppPath, "tools", "DOSPVR.exe");
            pvr2pngPath = Path.Combine(currentAppPath, "tools", "Pvr2Png.exe");

            builder = new DiscUtils.Gdrom.GDromBuilder()
            {
                RawMode = true,
                TruncateData = true,
            };
            builder.ReportProgress += ProgressReport;

            TempFolder = Path.GetTempPath();

            this.Closing += MainWindow_Closing;

            DataContext = this;

        }

        private void MainWindow_Closing(object sender, System.ComponentModel.CancelEventArgs e)
        {
            if (isAborting)
            {
                e.Cancel = true;
                return;
            }

            if (IsBusy)
            {
                if (ItemList.Any(x => x.State == ItemListState.Extracting || x.State == ItemListState.Inserting))
                {
                    if (MessageBox.Show("There are items still being processed.\nWant to cancel remaining queued items?", "Confirm", MessageBoxButton.YesNo, MessageBoxImage.Question) == MessageBoxResult.Yes)
                    {
                        isAborting = true;
                        MessageBox.Show("Remaining queued items have been aborted.\nPlease wait for current processing to finish.", "Information");
                    }
                    e.Cancel = true;
                    return;
                }
                //Cts.Cancel();
            }

            if (builder != null)
                builder.ReportProgress -= ProgressReport;
        }

        private async Task CopyTextureFileAsync(string temppath, string newThumbnail)
        {
            var fileExtension = Path.GetExtension(newThumbnail).ToUpperInvariant();
            if (fileExtension == ".PVR")
            {
                File.Copy(newThumbnail, Path.Combine(temppath, DATAPATH, GDTEXFILENAME), true);
            }
            else if (fileExtension == ".PNG")
            {
                //var notOnSameFolder = Path.GetDirectoryName(newThumbnail) == Path.GetDirectoryName(gdifilefullpath);

                //copy file to temp
                var gdtexnew = Path.Combine(temppath, Path.ChangeExtension(GDTEXFILENAME, fileExtension));
                File.Copy(newThumbnail, gdtexnew, true);
                //newThumbnail = gdtexnew;

                //File.Copy("", "", true);
                using (var p = CreateProcess(dospvrPath))
                {
                    p.StartInfo.Arguments = $"\"{gdtexnew}\" -CF SMART -TW -OP \"{Path.Combine(temppath, DATAPATH)}\" -Q";
                    p.Start();
                    await Task.Run(() => p.WaitForExit());
                }

                deleteFileIfExists(Path.Combine(temppath, DATAPATH, Path.ChangeExtension(GDTEXFILENAME, ".BAK")));
            }
        }
        

        private async Task<bool> ContainsTextureAsync(string originGdiFilePath)
        {
            bool toReturn = false;

            using (var p = CreateProcess(gditoolsPath))
            {
                p.StartInfo.Arguments = $"-i \"{originGdiFilePath}\" -l --silent";

                p.StartInfo.RedirectStandardOutput = true;
                p.StartInfo.RedirectStandardError = true;

                p.OutputDataReceived += (ss, ee) => { if (ee.Data?.ToUpperInvariant() == $@"/{GDTEXFILENAME}") toReturn = true; };
                p.ErrorDataReceived += (ss, ee) => { Debug.WriteLine(ee.Data); };

                p.Start();

                p.BeginOutputReadLine();
                p.BeginErrorReadLine();

                await Task.Run(() => p.WaitForExit());
                //gditools.exe -i "%~nx1\%GDI%" --data-folder "..\%~nx1 Extracted" -b "..\%~nx1 Extracted\bootsector\IP.BIN" --extract-all --silent
            }

            return toReturn;
        }

        private async Task ExtractAsync(string originGdiFilePath, string destinyFolderPath, bool textureOnly)
        {
            using (var p = CreateProcess(gditoolsPath))
            {
                if (textureOnly)
                    p.StartInfo.Arguments = $"-i \"{originGdiFilePath}\" -o \"{destinyFolderPath}\" -e 0GDTEX.PVR";
                else
                    p.StartInfo.Arguments = $"-i \"{originGdiFilePath}\" -o \"{destinyFolderPath}\" --data-folder {DATAPATH} -b {IPBINPATH} --extract-all --silent";

                p.StartInfo.RedirectStandardOutput = true;
                p.StartInfo.RedirectStandardError = true;

                p.OutputDataReceived += (ss, ee) => { Debug.WriteLine(ee.Data); };
                p.ErrorDataReceived += (ss, ee) => { Debug.WriteLine(ee.Data); };

                p.Start();

                p.BeginOutputReadLine();
                p.BeginErrorReadLine();

                await Task.Run(() => p.WaitForExit());
                //gditools.exe -i "%~nx1\%GDI%" --data-folder "..\%~nx1 Extracted" -b "..\%~nx1 Extracted\bootsector\IP.BIN" --extract-all --silent
            }
        }

        //private void rebuildByCommandLine(string tempfolderpath, string gdifilepath)
        //{
        //    using (var p = CreateProcess(buildgdiPath))
        //    {
        //        //p.StartInfo.Arguments = $" - data \"{Path.Combine(temppath, DATAPATH)}\" -ip \"{Path.Combine(temppath, IPBINPATH)}\" -output \"{gamefolder}\" -gdi \"{gdifilefullpath}\" -raw -truncate >NUL";
        //        p.StartInfo.Arguments = $"-data \"{DATAPATH}\" -ip \"{IPBINPATH}\" -output \"REBUILD\" -gdi \"{Path.Combine(tempfolderpath, gdifilepath)}\" -raw -truncate >NUL";
        //        p.StartInfo.WorkingDirectory = tempfolderpath;
        //        p.Start();
        //        p.WaitForExit();
        //        //-data "..\%~nx1" -ip "..\%~nx1\IP.BIN" -output "" -gdi "%GDI%" -raw -truncate >NUL
        //    }
        //}

        private List<string> GetCddaListFromGdi(string gdifilefullpath)
        {
            List<string> templist = GetFileListFromGdi(gdifilefullpath);
            for (int i = 0; i < templist.Count; i++)
                templist[i] = templist[i].ToUpperInvariant();


            string track04 = templist.FirstOrDefault(x => x == "TRACK04.RAW");
            if (track04 == null)
                return null;

            var i2 = templist.IndexOf(track04);
            return templist.Skip(i2).Where(x => Path.GetExtension(x) == ".RAW").Select(x => Path.Combine(Path.GetDirectoryName(gdifilefullpath), x)).ToList();
        }


        private async Task RebuildByDllAsync(string temppath, string gdifilefullpath)
        {
            var rebuildp = Path.Combine(temppath, REBUILDPATH);
            if (!Directory.Exists(rebuildp))
                Directory.CreateDirectory(rebuildp);

            var newgdipath = Path.Combine(rebuildp, Path.GetFileName(gdifilefullpath));

            //copiar gdi original para destino
            File.Copy(gdifilefullpath, newgdipath, true);


            //verificar se tem arquivos audio e criar tracklist
            var cddaList = GetCddaListFromGdi(gdifilefullpath);

            //var updatetDiscTracks = builder.BuildGDROM(Path.Combine(temppath, DATAPATH), Path.Combine(temppath, IPBINPATH), cddaList, rebuildp);
            var updatetDiscTracks = await Task.Run(() => builder.BuildGDROM(Path.Combine(temppath, DATAPATH), Path.Combine(temppath, IPBINPATH), cddaList, rebuildp));

            builder.UpdateGdiFile(updatetDiscTracks, newgdipath);

            //var RebuildFiles = new List<DiscUtils.Gdrom.DiscTrack>();
            
            List<string> RebuildGDIFiles = GetFileListFromGdi(newgdipath);

            var filesToKeep = new List<string>();
            foreach (var rebuildfile in RebuildGDIFiles)
                if (!File.Exists(Path.Combine(rebuildp, rebuildfile)))
                    filesToKeep.Add(rebuildfile.ToUpperInvariant());

            //apagar na pasta origem, tudo, menos os filestokeep
            var ddd = Path.GetDirectoryName(gdifilefullpath);
            var fff = Directory.GetFiles(ddd);
            foreach (var item in fff)
                if (!filesToKeep.Contains(Path.GetFileName(item).ToUpperInvariant()))
                    deleteFileIfExists(item);

            foreach (var item in Directory.GetFiles(rebuildp))
                await Task.Run(() => File.Move(item, Path.Combine(ddd, Path.GetFileName(item))));

            
            //mover arquivos da pasta REBUILD para pasta origem
            //apagar tudo da pasta temp
        }

        private static List<string> GetFileListFromGdi(string gdiFilePath)
        {
            var toReturn = new List<string>();
            Regex reg = new Regex(@"([\d]+) ([\d]+) ([\d]+) ([\d]+) ([\d\w]+.[\d\w]+) [\d]+");
            uint totalFilesGdi = 0;
            string line;
            using (StreamReader file = new StreamReader(gdiFilePath))
            {
                while ((line = file.ReadLine()) != null)
                {
                    if (totalFilesGdi == 0)
                    {
                        if (!uint.TryParse(line, out totalFilesGdi))
                            throw new Exception();
                    }
                    else
                    {
                        var m = reg.Match(line);
                        if (m.Success)
                            toReturn.Add(m.Groups[5].Value);
                    }
                }
            }
            if (totalFilesGdi != toReturn.Count)
                throw new Exception();

            return toReturn;
        }

        private Process CreateProcess(string executablePath)
        {
            var p = new Process();
            p.StartInfo.CreateNoWindow = true;
            p.StartInfo.UseShellExecute = false;
            p.StartInfo.FileName = executablePath;
            return p;
        }

        private void ProgressReport(int amount)
        {
            System.Diagnostics.Debug.WriteLine(amount);
            if (amount % 10 != 0)
                return;
            //Console.Write('.');
        }


        private async void DataGrid_Drop(object sender, DragEventArgs e)
        {
            try
            {
                IsBusy = true;

                if (e.Data.GetDataPresent(DataFormats.FileDrop, true))
                {
                    string[] filenames = e.Data.GetData(DataFormats.FileDrop, true) as string[];

                    foreach (string filename in filenames)
                    {
                        var gdiFiles = Directory.GetFiles(filename, gdiFilter);
                        if (gdiFiles != null)
                        {
                            EnsureHasOnlyOneGdiFile(filename, gdiFiles);

                            ItemList.Add(new GDIItem
                            {
                                Path = filename,
                                GdiFile = Path.GetFileName(gdiFiles[0])
                            });
                        }
                    }
                }
            }
            catch (GdiException ex)
            {
                await Task.Run(() => MessageBox.Show(ex.Message, "Warning", MessageBoxButton.OK, MessageBoxImage.Warning));
            }
            catch (TaskCanceledException)
            {
                ItemList.Clear();
            }
            catch (Exception ex)
            {
                await Task.Run(() => MessageBox.Show(ex.Message, "Error", MessageBoxButton.OK, MessageBoxImage.Error));
            }
            finally
            {
                IsBusy = false;
            }
        }


        private void ButtonAbout_Click(object sender, RoutedEventArgs e)
        {
            IsBusy = true;

            var helptext = @"Uses gditools by FamilyGuy, DiscUtils by Sappharad,
DOSPVR by David Harvey, Pvr2Png Gzav

*** How to use ***

- Configure Settings (optional).
- Drag and Drop Folders into List.
- Click the Extract or Insert button.


*** Extract ***
- Will extract the pvr file from gdi and convert it to png.
  Both files will be saved in same folder as gdi.

*** Insert ***

- Will convert png files to pvr and insert/rebuild the gdi,
  Overriding original files.";

            MessageBox.Show(helptext, $"{this.Title} - by Sonik", MessageBoxButton.OK, MessageBoxImage.Information);
            IsBusy = false;
        }

        private void ButtonFolder_Click(object sender, RoutedEventArgs e)
        {
            var btn = (Button)sender;

            using (var dialog = new System.Windows.Forms.FolderBrowserDialog())
            {
                if ((string)btn.CommandParameter == nameof(TempFolder) && !string.IsNullOrEmpty(TempFolder))
                    dialog.SelectedPath = TempFolder;

                if (dialog.ShowDialog() == System.Windows.Forms.DialogResult.OK)
                    TempFolder = dialog.SelectedPath;
            }
        }

        private void EnsureHasOnlyOneGdiFile(string fileName, string[] gdiFiles)
        {
            if (gdiFiles.Length == 0)
                throw new GdiException($"No GDI file in folder:\n{Path.GetFileName(fileName)}");
            else if (gdiFiles.Length > 1)
                throw new GdiException($"More than one GDI file in folder:\n{Path.GetFileName(fileName)}");
        }

        protected void raiseEvent([CallerMemberName] string propertyName = "")
        {
            PropertyChanged?.Invoke(this, new PropertyChangedEventArgs(propertyName));
        }


        private async Task ConvertPvr2PngAsync(string pvrPath, string pngPath)
        {
            using (var p = CreateProcess(pvr2pngPath))
            {
                //pvr2png <source> [<destination>] [-i] [-q] [-?]
                p.StartInfo.Arguments = $"\"{pvrPath}\" \"{pngPath}\"";

                p.StartInfo.RedirectStandardOutput = true;
                p.StartInfo.RedirectStandardError = true;

                p.OutputDataReceived += (ss, ee) => { Debug.WriteLine(ee.Data); };
                p.ErrorDataReceived += (ss, ee) => { Debug.WriteLine(ee.Data); };

                p.Start();

                p.BeginOutputReadLine();
                p.BeginErrorReadLine();

                await Task.Run(() => p.WaitForExit());
            }
        }

        private void deleteFileIfExists(string filePath)
        {
            if (File.Exists(filePath))
                File.Delete(filePath);
        }

        private async void ButtonExtractImages_Click(object sender, RoutedEventArgs e)
        {
            GDIItem item = null;
            foreach (var itemx in ItemList)
                itemx.State = ItemListState.Queued;

            GDIItem[] queuedItems = ItemList.ToArray();

            if (!queuedItems.Any())
                return;

            IsBusy = true;

            try
            {
                for (int i = 0; i < queuedItems.Length; i++)
                {
                    if (isAborting)
                        break;

                    item = queuedItems[i];

                    var gamefolder = item.Path;

                    var gdiFiles = Directory.GetFiles(gamefolder, gdiFilter);
                    if (gdiFiles != null)
                        EnsureHasOnlyOneGdiFile(gamefolder, gdiFiles);

                    var gdifilefullpath = Path.Combine(item.Path, item.GdiFile);


                    if (!await ContainsTextureAsync(gdifilefullpath))
                    {
                        item.State = ItemListState.Skipped;
                        continue;
                    }

                    var gdtexfilename = Path.Combine(item.Path, GDTEXFILENAME);
                    var gdtexbackupfilename = Path.Combine(item.Path, "0GDTEX_EXTRACTED.PVR");
                    var pngfilename = Path.Combine(item.Path, Path.ChangeExtension(GDTEXFILENAME, ".PNG"));


                    deleteFileIfExists(gdtexfilename);
                    deleteFileIfExists(gdtexbackupfilename);
                    deleteFileIfExists(pngfilename);

                    try
                    {
                        item.State = ItemListState.Extracting;
                        await ExtractAsync(gdifilefullpath, item.Path, true);

                        if (File.Exists(gdtexfilename))
                        {
                            File.Move(gdtexfilename, gdtexbackupfilename);

                            //convert to png
                            await ConvertPvr2PngAsync(gdtexbackupfilename, pngfilename);
                        }

                        item.State = ItemListState.Done;
                    }
                    catch (Exception)
                    {
                        throw;
                    }
                    finally
                    {
                        deleteFileIfExists(gdtexfilename);
                    }

                }

                MessageBox.Show("Done!", "Sucess", MessageBoxButton.OK, MessageBoxImage.Information);

            }
            catch (Exception ex)
            {
                if (item != null)
                    item.State = ItemListState.Error;

                MessageBox.Show(ex.Message, "Error", MessageBoxButton.OK, MessageBoxImage.Error);
            }
            finally
            {
                IsBusy = false;
                isAborting = false;
            }

        }


        private async void ButtonInsertImages_Click(object sender, RoutedEventArgs e)
        {
            GDIItem item = null;
            foreach (var itemx in ItemList)
                itemx.State = ItemListState.Queued;

            GDIItem[] queuedItems = ItemList.ToArray();

            if (!queuedItems.Any())
                return;

            if (MessageBox.Show("This will override files in folders.\nConfirm?", "Confirm", MessageBoxButton.YesNo, MessageBoxImage.Question) != MessageBoxResult.Yes)
                return;


            IsBusy = true;

            string tempPath = null;

            try
            {
                tempPath = Path.Combine(TempFolder, Guid.NewGuid().ToString());

                for (int i = 0; i < queuedItems.Length; i++)
                {
                    if (isAborting)
                        break;

                    if (!Directory.Exists(tempPath))
                        Directory.CreateDirectory(tempPath);

                    item = queuedItems[i];

                    var gamefolder = item.Path;

                    var gdiFiles = Directory.GetFiles(gamefolder, gdiFilter);
                    if (gdiFiles != null)
                        EnsureHasOnlyOneGdiFile(gamefolder, gdiFiles);

                    var gdifilefullpath = Path.Combine(item.Path, item.GdiFile);

                    var newThumbnail = Path.Combine(item.Path, Path.ChangeExtension(GDTEXFILENAME, ".PNG"));

                    //check for png
                    if (!File.Exists(newThumbnail))
                    {
                        item.State = ItemListState.Skipped;
                        continue;
                    }


                    //extract gdi
                    item.State = ItemListState.Extracting;
                    await ExtractAsync(gdifilefullpath, tempPath, false);

                    //convert png to pvr and or
                    //copy pvr file to extracted folder
                    await CopyTextureFileAsync(tempPath, newThumbnail);

                    //rebuild gdi
                    item.State = ItemListState.Inserting;
                    await RebuildByDllAsync(tempPath, gdifilefullpath);

                    if (tempPath != null && Directory.Exists(tempPath))
                        Directory.Delete(tempPath, true);

                    item.State = ItemListState.Done;
                }

                MessageBox.Show("Done!", "Sucess", MessageBoxButton.OK, MessageBoxImage.Information);
            }
            catch (Exception ex)
            {
                if (item != null)
                    item.State = ItemListState.Error;

                MessageBox.Show(ex.Message, "Error", MessageBoxButton.OK, MessageBoxImage.Error);
            }
            finally
            {
                //delete temporary files from temppath
                try
                {
                    if (tempPath != null && Directory.Exists(tempPath))
                        Directory.Delete(tempPath, true);
                }
                catch { }
                IsBusy = false;
            }
        }
    }

    public class GdiException : Exception
    {
        public GdiException(string message) : base(message) { }
    }

    public enum ItemListState
    {
        Queued,
        Extracting,
        Inserting,
        Skipped,
        Error,
        Done
    }

    public class GDIItem : INotifyPropertyChanged
    {
        //public bool IsDirectory { get; set; }
        public string Path { get; set; }
        public string GdiFile { get; set; }

        private ItemListState _State;
        public ItemListState State
        {
            get { return _State; }
            set { _State = value; PropertyChanged?.Invoke(this, new PropertyChangedEventArgs(nameof(State))); }
        }

        public event PropertyChangedEventHandler PropertyChanged;
    }

}
